use super::clock::clock;
use super::comment::comment;
use super::diary_sexp::diary_sexp;
use super::drawer::drawer;
use super::dynamic_block::dynamic_block;
use super::element::Element;
use super::fixed_width_area::fixed_width_area;
use super::footnote_definition::footnote_definition;
use super::greater_block::greater_block;
use super::horizontal_rule::horizontal_rule;
use super::keyword::keyword;
use super::lesser_block::comment_block;
use super::lesser_block::example_block;
use super::lesser_block::export_block;
use super::lesser_block::src_block;
use super::lesser_block::verse_block;
use super::paragraph::paragraph;
use super::plain_list::plain_list;
use super::util::start_of_line;
use super::Context;
use crate::error::Res;
use crate::parser::parser_with_context::parser_with_context;
use crate::parser::table::org_mode_table;
use nom::branch::alt;
use nom::combinator::map;
use nom::multi::many0;

#[tracing::instrument(ret, level = "debug")]
pub fn element<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, Element<'s>> {
    let non_paragraph_matcher = parser_with_context!(non_paragraph_element)(context);
    let paragraph_matcher = parser_with_context!(paragraph_element)(context);

    alt((non_paragraph_matcher, paragraph_matcher))(input)
}

pub fn paragraph_element<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, Element<'s>> {
    let paragraph_matcher = parser_with_context!(paragraph)(context);
    let keyword_matcher = parser_with_context!(keyword)(context);
    let (remaining, affiliated_keywords) = many0(keyword_matcher)(input)?;
    map(paragraph_matcher, Element::Paragraph)(remaining)
}
pub fn non_paragraph_element<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, Element<'s>> {
    let plain_list_matcher = parser_with_context!(plain_list)(context);
    let greater_block_matcher = parser_with_context!(greater_block)(context);
    let dynamic_block_matcher = parser_with_context!(dynamic_block)(context);
    let footnote_definition_matcher = parser_with_context!(footnote_definition)(context);
    let comment_matcher = parser_with_context!(comment)(context);
    let drawer_matcher = parser_with_context!(drawer)(context);
    let table_matcher = parser_with_context!(org_mode_table)(context);
    let verse_block_matcher = parser_with_context!(verse_block)(context);
    let comment_block_matcher = parser_with_context!(comment_block)(context);
    let example_block_matcher = parser_with_context!(example_block)(context);
    let export_block_matcher = parser_with_context!(export_block)(context);
    let src_block_matcher = parser_with_context!(src_block)(context);
    let clock_matcher = parser_with_context!(clock)(context);
    let diary_sexp_matcher = parser_with_context!(diary_sexp)(context);
    let fixed_width_area_matcher = parser_with_context!(fixed_width_area)(context);
    let horizontal_rule_matcher = parser_with_context!(horizontal_rule)(context);
    let keyword_matcher = parser_with_context!(keyword)(context);
    let (remaining, mut affiliated_keywords) = many0(keyword_matcher)(input)?;
    let (remaining, element) = match alt((
        map(plain_list_matcher, Element::PlainList),
        map(greater_block_matcher, Element::GreaterBlock),
        map(dynamic_block_matcher, Element::DynamicBlock),
        map(footnote_definition_matcher, Element::FootnoteDefinition),
        map(comment_matcher, Element::Comment),
        map(drawer_matcher, Element::Drawer),
        map(table_matcher, Element::Table),
        map(verse_block_matcher, Element::VerseBlock),
        map(comment_block_matcher, Element::CommentBlock),
        map(example_block_matcher, Element::ExampleBlock),
        map(export_block_matcher, Element::ExportBlock),
        map(src_block_matcher, Element::SrcBlock),
        map(clock_matcher, Element::Clock),
        map(diary_sexp_matcher, Element::DiarySexp),
        map(fixed_width_area_matcher, Element::FixedWidthArea),
        map(horizontal_rule_matcher, Element::HorizontalRule),
    ))(remaining)
    {
        the_ok @ Ok(_) => the_ok,
        Err(_) => {
            affiliated_keywords.clear();
            map(keyword_matcher, Element::Keyword)(input)
        }
    }?;
    Ok((remaining, element))
}
